// MQL4&5-code

//   MT4  MT5  .
// https://www.mql5.com/ru/code/16006

//  mqh-   #include      MQL5 (MT5-Hedge)   ,   MQL4.
// ..    ()   MQL4.      
// MQL5- .  ,  MQL5-    .
//       .   !

//   MQL4 -> MQL5      .
//        (     MT4  #property strict):

// #include <MT4Orders.mqh> //   #include <Expert/Expert.mqh>,    

//   (  )   MQL5-,    MT5-   MT4-,    .

//      ,         " "
//  ,  ,   ..

//      .

//      ,  MQL4-    MT5  ,  .

//   :
//   CloseBy- -     . ,  ,  .
//    TP  SL   -    (build 1470) MQL5    .
//    DEAL_ENTRY_INOUT  DEAL_ENTRY_OUT_BY .

// :
//    MT4 OrderSelect   SELECT_BY_TICKET     MODE_TRADES/MODE_HISTORY,
//   .. "     ".
//    MT5    ,
//    OrderSelect   SELECT_BY_TICKET       :
//     MODE_TRADES:    >   >  >  
//     MODE_HISTORY:  >   >   >  
//
// , OrderSelect   SELECT_BY_TICKET  MT5    ( )    ,    MT4.
//
//   OrdersTotal()   ,      MT5-.

//  :
// 03.08.2016:
//    -      -.
// 29.09.2016:
//   Add:     (SYMBOL_TRADE_EXECUTION_EXCHANGE). ,   - Netto( Hedge)-mode.
//   Add:   "  #include <Trade/Trade.mqh>,    "
//          "  #include <Expert/Expert.mqh>,    ".
//   Fix: OrderSend -   ,   .
// 13.11.2016:
//   Add:   OrderSend, OrderModify, OrderClose, OrderDelete    (-  ) -   MT4.
//              MT4ORDERS::OrderSend_MaxPause  .     MT5 ~1 .
//        -      . MT4ORDERS::OrderSend_MaxPause = 0 -  .
//   Add:   SlipPage (OrderSend, OrderClose)    -   Instant-,
//                    - ENUM_ORDER_TYPE_FILLING:
//        ORDER_FILLING_FOK, ORDER_FILLING_IOC  ORDER_FILLING_RETURN.
//                       .
//        :
//          OrderSend(Symb, Type, Lots, Price, ORDER_FILLING_FOK, SL, TP) -       ORDER_FILLING_FOK
//          OrderSend(Symb, Type, Lots, Price, ORDER_FILLING_IOC, SL, TP) -       ORDER_FILLING_IOC
//          OrderClose(Ticket, Lots, Price, ORDER_FILLING_RETURN) -   -    ORDER_FILLING_RETURN
//   Add: OrdersHistoryTotal()  OrderSelect(Pos, SELECT_BY_POS, MODE_HISTORY)  -   .
//             .
// 08.02.2017:
//   Add:  MT4ORDERS::LastTradeRequest  MT4ORDERS::LastTradeResult    MT5-OrderSend.
// 14.06.2017:
//   Add:      SL/TP   (  OrderClose).
//   Add: MagicNumber    long - 8  (  int - 4 ).
//   Add:   OrderSend, OrderClose  OrderModify    ( )   INT_MAX,   
//          MT5- (MT4ORDERS::LastTradeRequest),     .      MT5-,
//             MT4ORDERS::LastTradeCheckResult.
//            OrderModify  OrderClose  true,  - false.
//        OrderSend  0   ,  - -1.
//
//                INT_MIN,      MT5- 
//         (    INT_MAX)   .
//   Add:    MQL4- : OrderSendAsync, OrderModifyAsync, OrderCloseAsync, OrderDeleteAsync.
//          Result.request_id   ,  - 0.

#ifdef __MQL5__
#ifndef __MT4ORDERS__

#define __MT4ORDERS__

struct MT4_ORDER
{
  int Ticket;
  int Type;

  double Lots;

  string Symbol;
  string Comment;

  double OpenPrice;
  datetime OpenTime;

  double StopLoss;
  double TakeProfit;

  double ClosePrice;
  datetime CloseTime;

  datetime Expiration;

  long MagicNumber;

  double Profit;

  double Commission;
  double Swap;

  string ToString( void ) const
  {
    static const string Types[] = {"buy", "sell", "buy limit", "sell limit", "buy stop", "sell stop", "balance"};
    const int digits = (int)::SymbolInfoInteger(this.Symbol, SYMBOL_DIGITS);

    return("#" + (string)this.Ticket + " " +
           (string)this.OpenTime + " " +
           ((this.Type < ::ArraySize(Types)) ? Types[this.Type] : "unknown") + " " +
           ::DoubleToString(this.Lots, 2) + " " +
           this.Symbol + " " +
           ::DoubleToString(this.OpenPrice, digits) + " " +
           ::DoubleToString(this.StopLoss, digits) + " " +
           ::DoubleToString(this.TakeProfit, digits) + " " +
           ((this.CloseTime > 0) ? ((string)this.CloseTime + " ") : "") +
           ::DoubleToString(this.ClosePrice, digits) + " " +
           ::DoubleToString(this.Commission, 2) + " " +
           ::DoubleToString(this.Swap, 2) + " " +
           ::DoubleToString(this.Profit, 2) + " " +
           ((this.Comment == "") ? "" : (this.Comment + " ")) +
           (string)this.MagicNumber +
           (((this.Expiration > 0) ? (" expiration " + (string)this.Expiration): "")));
  }
};

#define RESERVE_SIZE 1000
#define DAY (24 * 3600)
#define HISTORY_PAUSE (MT4HISTORY::IsTester ? 0 : 5)

class MT4HISTORY
{
private:
  static const bool MT4HISTORY::IsTester;

  static ulong TicketDeal;
  static ulong TicketOrder;

  long Tickets[];
  uint Amount;

  datetime LastTime;

  int LastTotalDeals;
  int LastTotalOrders;

  datetime LastInitTime;

#define GETNEXTPOS_FUNCTION(NAME)                                  \
  static int GetNextPosMT4##NAME( int iPos )                       \
  {                                                                \
    const int Total = ::History##NAME##sTotal();                   \
                                                                   \
    while (iPos < Total)                                           \
    {                                                              \
      MT4HISTORY::Ticket##NAME = ::History##NAME##GetTicket(iPos); \
                                                                   \
      if (MT4HISTORY::IsMT4##NAME(MT4HISTORY::Ticket##NAME))       \
        break;                                                     \
                                                                   \
      iPos++;                                                      \
    }                                                              \
                                                                   \
    return(iPos);                                                  \
  }

  GETNEXTPOS_FUNCTION(Order)
  GETNEXTPOS_FUNCTION(Deal)

#undef GETNEXTPOS_FUNCTION

  bool RefreshHistory( void )
  {
    bool Res = false;

    const datetime LastTimeCurrent = ::TimeCurrent();

    if ((!MT4HISTORY::IsTester) && (LastTimeCurrent >= this.LastInitTime + DAY))
    {
      this.LastTime = 0;

      this.LastTotalOrders = 0;
      this.LastTotalDeals = 0;

      this.Amount = 0;

      ::ArrayResize(this.Tickets, this.Amount, RESERVE_SIZE);

      this.LastInitTime = LastTimeCurrent;
    }

    if (::HistorySelect(this.LastTime, ::MathMax(LastTimeCurrent, this.LastTime) + DAY)) //  
    {
      const int TotalOrders = ::HistoryOrdersTotal();
      const int TotalDeals = ::HistoryDealsTotal();

      Res = ((TotalOrders > this.LastTotalOrders) || (TotalDeals > this.LastTotalDeals));

      if (Res)
      {
        int iOrder = MT4HISTORY::GetNextPosMT4Order(this.LastTotalOrders);
        int iDeal = MT4HISTORY::GetNextPosMT4Deal(this.LastTotalDeals);

        long TimeOrder = (iOrder < TotalOrders) ? ::HistoryOrderGetInteger(MT4HISTORY::TicketOrder, ORDER_TIME_DONE_MSC) : LONG_MAX;
        long TimeDeal = (iDeal < TotalDeals) ? ::HistoryDealGetInteger(MT4HISTORY::TicketDeal, DEAL_TIME_MSC) : LONG_MAX;

        while ((iDeal < TotalDeals) || (iOrder < TotalOrders))
          if (TimeOrder < TimeDeal)
          {
            this.Amount = ::ArrayResize(this.Tickets, this.Amount + 1, RESERVE_SIZE);

            this.Tickets[this.Amount - 1] = -(long)MT4HISTORY::TicketOrder;

            iOrder = MT4HISTORY::GetNextPosMT4Order(iOrder + 1);

            TimeOrder = (iOrder < TotalOrders) ? ::HistoryOrderGetInteger(MT4HISTORY::TicketOrder, ORDER_TIME_DONE_MSC) : LONG_MAX;
          }
          else
          {
            this.Amount = ::ArrayResize(this.Tickets, this.Amount + 1, RESERVE_SIZE);

            this.Tickets[this.Amount - 1] = (long)MT4HISTORY::TicketDeal;

            iDeal = MT4HISTORY::GetNextPosMT4Deal(iDeal + 1);

            TimeDeal = (iDeal < TotalDeals) ? ::HistoryDealGetInteger(MT4HISTORY::TicketDeal, DEAL_TIME_MSC) : LONG_MAX;
          }

        TimeOrder = (TotalOrders > this.LastTotalOrders) ? ::HistoryOrderGetInteger(MT4HISTORY::TicketOrder, ORDER_TIME_DONE_MSC) : 0;
        TimeDeal = (TotalDeals > this.LastTotalDeals) ? ::HistoryDealGetInteger(MT4HISTORY::TicketDeal, DEAL_TIME_MSC) : 0;

        const long MaxTime = ::MathMax(TimeOrder, TimeDeal);

        this.LastTotalOrders = (TimeOrder == MaxTime) ? 1 : 0;
        this.LastTotalDeals = (TimeDeal == MaxTime) ? 1 : 0;

        if ((LastTimeCurrent - HISTORY_PAUSE) * 1000 > MaxTime)
        {
          this.LastTime = LastTimeCurrent - HISTORY_PAUSE;

          this.LastTotalOrders = 0;
          this.LastTotalDeals = 0;
        }
        else
        {
          this.LastTime = (datetime)(MaxTime / 1000);

          if (TimeOrder == MaxTime)
            for (int i = TotalOrders - 2; i >= 0; i--)
            {
              if (this.LastTime > ::HistoryOrderGetInteger(::HistoryOrderGetTicket(i), ORDER_TIME_DONE))
                break;

              this.LastTotalOrders++;
            }

          if (TimeDeal == MaxTime)
            for (int i = TotalDeals - 2; i >= 0; i--)
            {
              if (this.LastTime > ::HistoryDealGetInteger(::HistoryDealGetTicket(i), DEAL_TIME))
                break;

              this.LastTotalDeals++;
            }
        }
      }
      else if (LastTimeCurrent - HISTORY_PAUSE > this.LastTime)
      {
        this.LastTime = LastTimeCurrent - HISTORY_PAUSE;

        this.LastTotalOrders = 0;
        this.LastTotalDeals = 0;
      }
    }

    return(Res);
  }

public:
  static bool IsMT4Deal( const ulong Ticket )
  {
    const ENUM_DEAL_TYPE DealType = (ENUM_DEAL_TYPE)::HistoryDealGetInteger(Ticket, DEAL_TYPE);

    return(((DealType != DEAL_TYPE_BUY) && (DealType != DEAL_TYPE_SELL)) ||
           ((ENUM_DEAL_ENTRY)::HistoryDealGetInteger(Ticket, DEAL_ENTRY) == DEAL_ENTRY_OUT));
  }

  static bool IsMT4Order( const ulong Ticket )
  {
    return((::HistoryOrderGetDouble(Ticket, ORDER_VOLUME_CURRENT) > 0) || (::HistoryOrderGetInteger(Ticket, ORDER_POSITION_ID) == 0));
  }

  MT4HISTORY( void ) : Amount(::ArrayResize(this.Tickets, 0, RESERVE_SIZE)),
                       LastTime(0), LastTotalDeals(0), LastTotalOrders(0), LastInitTime(0)
  {
//    this.RefreshHistory(); //    ,   .
  }

  int GetAmount( void )
  {
    this.RefreshHistory();

    return((int)this.Amount);
  }

  long operator []( const uint Pos )
  {
    long Res = 0;

    if (Pos >= this.Amount)
    {
      this.RefreshHistory();

      if (Pos < this.Amount)
        Res = this.Tickets[Pos];
    }
    else
      Res = this.Tickets[Pos];

    return(Res);
  }
};

static const bool MT4HISTORY::IsTester = (::MQLInfoInteger(MQL_TESTER) || ::MQLInfoInteger(MQL_OPTIMIZATION) ||
                                          ::MQLInfoInteger(MQL_VISUAL_MODE) || ::MQLInfoInteger(MQL_FRAME_MODE));

static ulong MT4HISTORY::TicketDeal = 0;
static ulong MT4HISTORY::TicketOrder = 0;

#undef HISTORY_PAUSE
#undef DAY
#undef RESERVE_SIZE

#define OP_BUY ORDER_TYPE_BUY
#define OP_SELL ORDER_TYPE_SELL
#define OP_BUYLIMIT ORDER_TYPE_BUY_LIMIT
#define OP_SELLLIMIT ORDER_TYPE_SELL_LIMIT
#define OP_BUYSTOP ORDER_TYPE_BUY_STOP
#define OP_SELLSTOP ORDER_TYPE_SELL_STOP
#define OP_BALANCE 6

#define SELECT_BY_POS 0
#define SELECT_BY_TICKET 1

#define MODE_TRADES 0
#define MODE_HISTORY 1

class MT4ORDERS
{
private:
  static MT4_ORDER Order;
  static MT4HISTORY History;

  static const bool MT4ORDERS::IsTester;
  static const bool MT4ORDERS::IsHedging;

  static bool OrderSendBug;

  static ulong GetPositionDealIn( const ulong PositionIdentifier = -1 ) // 0 - , ..     
  {
    ulong Ticket = 0;

    if (PositionIdentifier == -1)
    {
      const datetime PosTime = (datetime)::PositionGetInteger(POSITION_TIME);

      if (::HistorySelect(PosTime, PosTime))
      {
        const int Total = ::HistoryDealsTotal();
        const int MyPositionIdentifier = (int)::PositionGetInteger(POSITION_IDENTIFIER);

        for (int i = 0; i < Total; i++)
        {
          const ulong TicketDeal = ::HistoryDealGetTicket(i);

          if ((::HistoryDealGetInteger(TicketDeal, DEAL_POSITION_ID) == MyPositionIdentifier) /*&&
              ((ENUM_DEAL_ENTRY)::HistoryDealGetInteger(TicketDeal, DEAL_ENTRY) == DEAL_ENTRY_IN) */) //      DEAL_ENTRY_IN
          {
            Ticket = TicketDeal;

            break;
          }
        }
      }
    }
    else if ((PositionIdentifier != 0) && // PositionIdentifier    
             ::HistorySelectByPosition(PositionIdentifier) && (::HistoryDealsTotal() > 1))
    {
      const int Tmp = ::HistoryDealsTotal();
      Ticket = ::HistoryDealGetTicket(0); //      DEAL_ENTRY_IN
      /*
      const int Total = ::HistoryDealsTotal();

      for (int i = 0; i < Total; i++)
      {
        const ulong TicketDeal = ::HistoryDealGetTicket(i);

        if (TicketDeal > 0)
          if ((ENUM_DEAL_ENTRY)::HistoryDealGetInteger(TicketDeal, DEAL_ENTRY) == DEAL_ENTRY_IN)
          {
            Ticket = TicketDeal;

            break;
          }
      } */
    }

    return(Ticket);
  }

  static void CheckPositionCommissionComment( void )
  {
    if (((MT4ORDERS::Order.Commission == DBL_MIN) || (MT4ORDERS::Order.Comment == NULL)) &&
        ::PositionSelectByTicket(MT4ORDERS::Order.Ticket)) // POSITION_VOLUME  
    {
      MT4ORDERS::Order.Commission = ::PositionGetDouble(POSITION_COMMISSION);
      MT4ORDERS::Order.Comment = ::PositionGetString(POSITION_COMMENT);

      if ((MT4ORDERS::Order.Commission == 0) || (MT4ORDERS::Order.Comment == ""))
      {
        const ulong Ticket = MT4ORDERS::GetPositionDealIn(); //  -    

        if (Ticket > 0)
        {
          if (MT4ORDERS::Order.Commission == 0)
          {
            const double LotsIn = ::HistoryDealGetDouble(Ticket, DEAL_VOLUME);

            if (LotsIn > 0)
              MT4ORDERS::Order.Commission = ::HistoryDealGetDouble(Ticket, DEAL_COMMISSION) * ::PositionGetDouble(POSITION_VOLUME) / LotsIn;
          }

          if (MT4ORDERS::Order.Comment == "")
            MT4ORDERS::Order.Comment = ::HistoryDealGetString(Ticket, DEAL_COMMENT);
        }
      }
    }

    return;
  }

  static void GetPositionData( void )
  {
    MT4ORDERS::Order.Ticket = (int)::PositionGetInteger(POSITION_TICKET);
    MT4ORDERS::Order.Type = (int)::PositionGetInteger(POSITION_TYPE);

    MT4ORDERS::Order.Lots = ::PositionGetDouble(POSITION_VOLUME);

    MT4ORDERS::Order.Symbol = ::PositionGetString(POSITION_SYMBOL);
    MT4ORDERS::Order.Comment = NULL; // MT4ORDERS::CheckPositionCommissionComment();

    MT4ORDERS::Order.OpenPrice = ::PositionGetDouble(POSITION_PRICE_OPEN);
    MT4ORDERS::Order.OpenTime = (datetime)::PositionGetInteger(POSITION_TIME);

    MT4ORDERS::Order.StopLoss = ::PositionGetDouble(POSITION_SL);
    MT4ORDERS::Order.TakeProfit = ::PositionGetDouble(POSITION_TP);

    MT4ORDERS::Order.ClosePrice = ::PositionGetDouble(POSITION_PRICE_CURRENT);
    MT4ORDERS::Order.CloseTime = 0;

    MT4ORDERS::Order.Expiration = 0;

    MT4ORDERS::Order.MagicNumber = ::PositionGetInteger(POSITION_MAGIC);

    MT4ORDERS::Order.Profit = ::PositionGetDouble(POSITION_PROFIT);

    MT4ORDERS::Order.Commission = DBL_MIN; // MT4ORDERS::CheckPositionCommissionComment();
    MT4ORDERS::Order.Swap = ::PositionGetDouble(POSITION_SWAP);

    return;
  }

  static void GetOrderData( void )
  {
    MT4ORDERS::Order.Ticket = (int)::OrderGetInteger(ORDER_TICKET);
    MT4ORDERS::Order.Type = (int)::OrderGetInteger(ORDER_TYPE);

    MT4ORDERS::Order.Lots = ::OrderGetDouble(ORDER_VOLUME_CURRENT);

    MT4ORDERS::Order.Symbol = ::OrderGetString(ORDER_SYMBOL);
    MT4ORDERS::Order.Comment = ::OrderGetString(ORDER_COMMENT);

    MT4ORDERS::Order.OpenPrice = ::OrderGetDouble(ORDER_PRICE_OPEN);
    MT4ORDERS::Order.OpenTime = (datetime)::OrderGetInteger(ORDER_TIME_SETUP);

    MT4ORDERS::Order.StopLoss = ::OrderGetDouble(ORDER_SL);
    MT4ORDERS::Order.TakeProfit = ::OrderGetDouble(ORDER_TP);

    MT4ORDERS::Order.ClosePrice = ::OrderGetDouble(ORDER_PRICE_CURRENT);
    MT4ORDERS::Order.CloseTime = (datetime)::OrderGetInteger(ORDER_TIME_DONE);

    MT4ORDERS::Order.Expiration = (datetime)::OrderGetInteger(ORDER_TIME_EXPIRATION);

    MT4ORDERS::Order.MagicNumber = ::OrderGetInteger(ORDER_MAGIC);

    MT4ORDERS::Order.Profit = 0;

    MT4ORDERS::Order.Commission = 0;
    MT4ORDERS::Order.Swap = 0;

    return;
  }

  static void GetHistoryOrderData( const ulong Ticket )
  {
    MT4ORDERS::Order.Ticket = (int)::HistoryOrderGetInteger(Ticket, ORDER_TICKET);
    MT4ORDERS::Order.Type = (int)::HistoryOrderGetInteger(Ticket, ORDER_TYPE);

    MT4ORDERS::Order.Lots = ::HistoryOrderGetDouble(Ticket, ORDER_VOLUME_CURRENT);

    if (MT4ORDERS::Order.Lots == 0)
      MT4ORDERS::Order.Lots = ::HistoryOrderGetDouble(Ticket, ORDER_VOLUME_INITIAL);

    MT4ORDERS::Order.Symbol = ::HistoryOrderGetString(Ticket, ORDER_SYMBOL);
    MT4ORDERS::Order.Comment = ::HistoryOrderGetString(Ticket, ORDER_COMMENT);

    MT4ORDERS::Order.OpenPrice = ::HistoryOrderGetDouble(Ticket, ORDER_PRICE_OPEN);
    MT4ORDERS::Order.OpenTime = (datetime)::HistoryOrderGetInteger(Ticket, ORDER_TIME_SETUP);

    MT4ORDERS::Order.StopLoss = ::HistoryOrderGetDouble(Ticket, ORDER_SL);
    MT4ORDERS::Order.TakeProfit = ::HistoryOrderGetDouble(Ticket, ORDER_TP);

    MT4ORDERS::Order.ClosePrice = 0; // ::HistoryOrderGetDouble(Ticket, ORDER_PRICE_CURRENT);
    MT4ORDERS::Order.CloseTime = (datetime)::HistoryOrderGetInteger(Ticket, ORDER_TIME_DONE);

    MT4ORDERS::Order.Expiration = (datetime)::HistoryOrderGetInteger(Ticket, ORDER_TIME_EXPIRATION);

    MT4ORDERS::Order.MagicNumber = ::HistoryOrderGetInteger(Ticket, ORDER_MAGIC);

    MT4ORDERS::Order.Profit = 0;

    MT4ORDERS::Order.Commission = 0;
    MT4ORDERS::Order.Swap = 0;

    return;
  }

  static void GetHistoryPositionData( const ulong Ticket )
  {
    MT4ORDERS::Order.Ticket = (int)::HistoryDealGetInteger(Ticket, DEAL_TICKET);
    MT4ORDERS::Order.Type = (int)::HistoryDealGetInteger(Ticket, DEAL_TYPE);

    if ((MT4ORDERS::Order.Type > OP_SELL))
      MT4ORDERS::Order.Type += (OP_BALANCE - OP_SELL - 1);
    else
      MT4ORDERS::Order.Type = 1 - MT4ORDERS::Order.Type;

    MT4ORDERS::Order.Lots = ::HistoryDealGetDouble(Ticket, DEAL_VOLUME);

    MT4ORDERS::Order.Symbol = ::HistoryDealGetString(Ticket, DEAL_SYMBOL);
    MT4ORDERS::Order.Comment = ::HistoryDealGetString(Ticket, DEAL_COMMENT);

    MT4ORDERS::Order.OpenPrice = ::HistoryDealGetDouble(Ticket, DEAL_PRICE);
    MT4ORDERS::Order.OpenTime = (datetime)::HistoryDealGetInteger(Ticket, DEAL_TIME);

    //  StopLoss  TakeProfit   - build 1368
    MT4ORDERS::Order.StopLoss = 0; // ::HistoryDealGetDouble(Ticket, DEAL_SL);
    MT4ORDERS::Order.TakeProfit = 0; // ::HistoryDealGetDouble(Ticket, DEAL_TP);

    MT4ORDERS::Order.ClosePrice = MT4ORDERS::Order.OpenPrice; // ::HistoryDealGetDouble(Ticket, DEAL_PRICE);
    MT4ORDERS::Order.CloseTime = MT4ORDERS::Order.OpenTime;   // (datetime)::HistoryDealGetInteger(Ticket, DEAL_TIME);;

    MT4ORDERS::Order.Expiration = 0;

    MT4ORDERS::Order.MagicNumber = ::HistoryDealGetInteger(Ticket, DEAL_MAGIC);

    MT4ORDERS::Order.Profit = ::HistoryDealGetDouble(Ticket, DEAL_PROFIT);

    MT4ORDERS::Order.Commission = ::HistoryDealGetDouble(Ticket, DEAL_COMMISSION);
    MT4ORDERS::Order.Swap = ::HistoryDealGetDouble(Ticket, DEAL_SWAP);

    const ulong OrderTicket = ::HistoryDealGetInteger(Ticket, DEAL_ORDER);
    const ulong OpenTicket = MT4ORDERS::GetPositionDealIn(::HistoryDealGetInteger(Ticket, DEAL_POSITION_ID));

    if (OpenTicket > 0)
    {
      //  -  : . OrderClose.  ,  GetPositionDealIn, - HistorySelectByPosition
      MT4ORDERS::Order.StopLoss = ::HistoryOrderGetDouble(OrderTicket, ORDER_TP);
      MT4ORDERS::Order.TakeProfit = ::HistoryOrderGetDouble(OrderTicket, ORDER_SL);

      MT4ORDERS::Order.OpenPrice = ::HistoryDealGetDouble(OpenTicket, DEAL_PRICE);
      MT4ORDERS::Order.OpenTime = (datetime)::HistoryDealGetInteger(OpenTicket, DEAL_TIME);

      const double OpenLots = ::HistoryDealGetDouble(OpenTicket, DEAL_VOLUME);

      if (OpenLots > 0)
        MT4ORDERS::Order.Commission += ::HistoryDealGetDouble(OpenTicket, DEAL_COMMISSION) * MT4ORDERS::Order.Lots / OpenLots;

      if (MT4ORDERS::Order.MagicNumber == 0)
        MT4ORDERS::Order.MagicNumber = ::HistoryDealGetInteger(OpenTicket, DEAL_MAGIC);

      if (MT4ORDERS::Order.Comment == "")
        MT4ORDERS::Order.Comment = ::HistoryDealGetString(OpenTicket, DEAL_COMMENT);
    }

    return;
  }

  static bool Waiting( const bool FlagInit = false )
  {
    static ulong StartTime = 0;

    const bool Res = FlagInit ? false : (::GetMicrosecondCount() - StartTime < MT4ORDERS::OrderSend_MaxPause);

    if (FlagInit)
    {
      StartTime = ::GetMicrosecondCount();

      MT4ORDERS::OrderSendBug = false;
    }
    else if (Res)
    {
      ::Sleep(0);

      MT4ORDERS::OrderSendBug = true;
    }

    return(Res);
  }

  static bool EqualPrices( const double Price1, const double Price2, const int digits)
  {
    return(::NormalizeDouble(Price1 - Price2, digits) == 0);
  }

  static bool HistoryDealSelect( MqlTradeResult &Result )
  {
    if ((Result.deal == 0) && (Result.order != 0))
    {
      if (::HistorySelectByPosition(::HistoryOrderGetInteger(Result.order, ORDER_POSITION_ID)))
        for (int i = ::HistoryDealsTotal() - 1; i >= 0; i--)
        {
          const ulong DealTicket = ::HistoryDealGetTicket(i);

          if (Result.order == ::HistoryDealGetInteger(DealTicket, DEAL_ORDER))
          {
            Result.deal = DealTicket;

            break;
          }
        }
    }

    return(::HistoryDealSelect(Result.deal));
  }

#define TMP_MT4ORDERS_BENCHMARK(A) \
  static ulong Max##A = 0;         \
                                   \
  if (Interval##A > Max##A)        \
  {                                \
    MT4ORDERS_BENCHMARK            \
                                   \
    Max##A = Interval##A;          \
  }

  static void OrderSend_Benchmark( const ulong Interval1, const ulong Interval2 = 0 )
  {
    #ifdef MT4ORDERS_BENCHMARK
      TMP_MT4ORDERS_BENCHMARK(1)
      TMP_MT4ORDERS_BENCHMARK(2)
    #endif // MT4ORDERS_BENCHMARK

    return;
  }

#undef TMP_MT4ORDERS_BENCHMARK

#define WHILE(A) while ((!(Res = (A))) && MT4ORDERS::Waiting())

  static bool OrderSend( const MqlTradeRequest &Request, MqlTradeResult &Result )
  {
    const ulong StartTime1 = MT4ORDERS::IsTester ? 0 : ::GetMicrosecondCount();

    bool Res = ::OrderSend(Request, Result);

    const ulong Interval1 = MT4ORDERS::IsTester ? 0 : (::GetMicrosecondCount() - StartTime1);

    const ulong StartTime2 = MT4ORDERS::IsTester ? 0 : ::GetMicrosecondCount();

    if (Res && !MT4ORDERS::IsTester && (Result.retcode < TRADE_RETCODE_ERROR) && (MT4ORDERS::OrderSend_MaxPause > 0))
    {
      Res = (Result.retcode == TRADE_RETCODE_DONE);
      MT4ORDERS::Waiting(true);

      if (Request.action == TRADE_ACTION_DEAL)
      {
        WHILE(::HistoryOrderSelect(Result.order))
          ;

        Res = Res && (((ENUM_ORDER_STATE)::HistoryOrderGetInteger(Result.order, ORDER_STATE) == ORDER_STATE_FILLED) ||
                      ((ENUM_ORDER_STATE)::HistoryOrderGetInteger(Result.order, ORDER_STATE) == ORDER_STATE_PARTIAL));

        if (Res)
          WHILE(MT4ORDERS::HistoryDealSelect(Result))
            ;
      }
      else if (Request.action == TRADE_ACTION_PENDING)
      {
        if (Res)
          WHILE(::OrderSelect(Result.order) && ((ENUM_ORDER_STATE)::OrderGetInteger(ORDER_STATE) == ORDER_STATE_PLACED))
            ;
        else
        {
          WHILE(::HistoryOrderSelect(Result.order))
            ;

          Res = false;
        }
      }
      else if (Request.action == TRADE_ACTION_SLTP)
      {
        if (Res)
        {
          bool EqualSL = false;
          bool EqualTP = false;

          const int digits = (int)::SymbolInfoInteger(Request.symbol, SYMBOL_DIGITS);

          if ((Request.position == 0) ? ::PositionSelect(Request.symbol) : ::PositionSelectByTicket(Request.position))
          {
            EqualSL = MT4ORDERS::EqualPrices(::PositionGetDouble(POSITION_SL), Request.sl, digits);
            EqualTP = MT4ORDERS::EqualPrices(::PositionGetDouble(POSITION_TP), Request.tp, digits);
          }

          WHILE((EqualSL && EqualTP))
            if ((Request.position == 0) ? ::PositionSelect(Request.symbol) : ::PositionSelectByTicket(Request.position))
            {
              EqualSL = MT4ORDERS::EqualPrices(::PositionGetDouble(POSITION_SL), Request.sl, digits);
              EqualTP = MT4ORDERS::EqualPrices(::PositionGetDouble(POSITION_TP), Request.tp, digits);
            }
        }
      }
      else if (Request.action == TRADE_ACTION_MODIFY)
      {
        if (Res)
        {
          bool EqualSL = false;
          bool EqualTP = false;
          bool EqualPrice = false;

          const int digits = (int)::SymbolInfoInteger(Request.symbol, SYMBOL_DIGITS);

          if (::OrderSelect(Result.order))
          {
            EqualSL = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_SL), Request.sl, digits);
            EqualTP = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_TP), Request.tp, digits);
            EqualPrice = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_PRICE_OPEN), Request.price, digits);
          }

          WHILE((EqualSL && EqualTP && EqualPrice))
            if (::OrderSelect(Result.order) && ((ENUM_ORDER_STATE)::OrderGetInteger(ORDER_STATE) != ORDER_STATE_REQUEST_MODIFY))
            {
              EqualSL = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_SL), Request.sl, digits);
              EqualTP = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_TP), Request.tp, digits);
              EqualPrice = MT4ORDERS::EqualPrices(::OrderGetDouble(ORDER_PRICE_OPEN), Request.price, digits);
            }
        }
      }
      else if (Request.action == TRADE_ACTION_REMOVE)
        if (Res)
          WHILE(::HistoryOrderSelect(Result.order))
            ;
    }

    const ulong Interval2 = ::GetMicrosecondCount() - StartTime2;

    Result.comment += " " + ::DoubleToString(Interval1 / 1000.0, 3) + " + " + ::DoubleToString(Interval2 / 1000.0, 3) + " ms";

    MT4ORDERS::OrderSend_Benchmark(Interval1, Interval2);

    if (MT4ORDERS::OrderSendBug)
    {
      Alert("BUG!!!!!!");

      //MT4ORDERS_BENCHMARK;
    }

    return(Res);
  }

#undef WHILE

  static ENUM_DAY_OF_WEEK GetDayOfWeek( const datetime time )
  {
    MqlDateTime sTime = {0};

    ::TimeToStruct(time, sTime);

    return((ENUM_DAY_OF_WEEK)sTime.day_of_week);
  }

  static bool SessionTrade( const string Symb )
  {
    datetime TimeNow = ::TimeCurrent();

    const ENUM_DAY_OF_WEEK DayOfWeek = MT4ORDERS::GetDayOfWeek(TimeNow);

    TimeNow %= 24 * 60 * 60;

    bool Res = false;
    datetime From, To;

    for (int i = 0; (!Res) && ::SymbolInfoSessionTrade(Symb, DayOfWeek, i, From, To); i++)
      Res = ((From <= TimeNow) && (TimeNow < To));

    return(Res);
  }

  static bool SymbolTrade( const string Symb )
  {
    MqlTick Tick;

    return(::SymbolInfoTick(Symb, Tick) ? ((Tick.bid != 0) && (Tick.ask != 0) && MT4ORDERS::SessionTrade(Symb) /* &&
           ((ENUM_SYMBOL_TRADE_MODE)::SymbolInfoInteger(Symb, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL) */) : false);
  }

  static bool NewOrderCheck( void )
  {
    return(::OrderCheck(MT4ORDERS::LastTradeRequest, MT4ORDERS::LastTradeCheckResult) &&
           (MT4ORDERS::IsTester || MT4ORDERS::SymbolTrade(MT4ORDERS::LastTradeRequest.symbol)));
  }

  static bool NewOrderSend( const int Check = 0 )
  {
    return((Check == INT_MAX) ? MT4ORDERS::NewOrderCheck() :
           (((Check != INT_MIN) || MT4ORDERS::NewOrderCheck()) && MT4ORDERS::OrderSend(MT4ORDERS::LastTradeRequest, MT4ORDERS::LastTradeResult) ? MT4ORDERS::LastTradeResult.retcode < TRADE_RETCODE_ERROR : false));
  }

  static bool ModifyPosition( const int Ticket, MqlTradeRequest &Request )
  {
    const bool Res = ::PositionSelectByTicket(Ticket);

    if (Res)
    {
      Request.action = TRADE_ACTION_SLTP;

      Request.position = Ticket;
      Request.symbol = ::PositionGetString(POSITION_SYMBOL); //     !
    }

    return(Res);
  }

  static ENUM_ORDER_TYPE_FILLING GetFilling( const string Symb, const uint Type = ORDER_FILLING_FOK )
  {
    static ENUM_ORDER_TYPE_FILLING Res = ORDER_FILLING_FOK;
    static string LastSymb = NULL;
    static uint LastType = ORDER_FILLING_FOK;

    if ((LastType != Type) || (LastSymb != Symb))
    {
      LastType = Type;
      LastSymb = Symb;

      const ENUM_SYMBOL_TRADE_EXECUTION ExeMode = (ENUM_SYMBOL_TRADE_EXECUTION)::SymbolInfoInteger(Symb, SYMBOL_TRADE_EXEMODE);
      const int FillingMode = (int)::SymbolInfoInteger(Symb, SYMBOL_FILLING_MODE);

      Res = (FillingMode == 0 || (Type >= ORDER_FILLING_RETURN) || ((FillingMode & (Type + 1)) != Type + 1)) ?
            (((ExeMode == SYMBOL_TRADE_EXECUTION_EXCHANGE) || (ExeMode == SYMBOL_TRADE_EXECUTION_INSTANT)) ?
             ORDER_FILLING_RETURN : ((FillingMode == SYMBOL_FILLING_IOC) ? ORDER_FILLING_IOC : ORDER_FILLING_FOK)) :
            (ENUM_ORDER_TYPE_FILLING)Type;
    }

    return(Res);

  }

  static ENUM_ORDER_TYPE_TIME GetExpirationType( const string Symb, uint Expiration = ORDER_TIME_GTC )
  {
    static ENUM_ORDER_TYPE_TIME Res = ORDER_TIME_GTC;
    static string LastSymb = NULL;
    static uint LastExpiration = ORDER_TIME_GTC;

    if ((LastExpiration != Expiration) || (LastSymb != Symb))
    {
      LastExpiration = Expiration;
      LastSymb = Symb;

      const int ExpirationMode = (int)::SymbolInfoInteger(Symb, SYMBOL_EXPIRATION_MODE);

      if ((Expiration > ORDER_TIME_SPECIFIED_DAY) || (((ExpirationMode >> Expiration) & 1) == 0))
      {
        if ((Expiration < ORDER_TIME_SPECIFIED) || (ExpirationMode < SYMBOL_EXPIRATION_SPECIFIED))
          Expiration = ORDER_TIME_GTC;
        else if (Expiration > ORDER_TIME_DAY)
          Expiration = ORDER_TIME_SPECIFIED;

        uint i = 1 << Expiration;

        while ((Expiration <= ORDER_TIME_SPECIFIED_DAY) && ((ExpirationMode & i) != i))
        {
          i <<= 1;
          Expiration++;
        }
      }

      Res = (ENUM_ORDER_TYPE_TIME)Expiration;
    }

    return(Res);
  }

  static bool ModifyOrder( const int Ticket, const double Price, const datetime Expiration, MqlTradeRequest &Request )
  {
    const bool Res = ::OrderSelect(Ticket);

    if (Res)
    {
      Request.action = TRADE_ACTION_MODIFY;
      Request.order = Ticket;

      Request.price = Price;

      Request.symbol = ::OrderGetString(ORDER_SYMBOL);

      // https://www.mql5.com/ru/forum/1111/page1817#comment_4087275
//      Request.type_filling = (ENUM_ORDER_TYPE_FILLING)::OrderGetInteger(ORDER_TYPE_FILLING);
      Request.type_filling = MT4ORDERS::GetFilling(Request.symbol);
      Request.type_time = MT4ORDERS::GetExpirationType(Request.symbol, (uint)Expiration);

      if (Expiration > ORDER_TIME_DAY)
        Request.expiration = Expiration;
    }

    return(Res);
  }

  static bool SelectByPosHistory( const int Index )
  {
    const int Ticket = (int)MT4ORDERS::History[Index];
    const bool Res = (Ticket > 0) ? ::HistoryDealSelect(Ticket) : ((Ticket < 0) ? ::HistoryOrderSelect(-Ticket) : false);

    if (Res)
    {
      if (Ticket > 0)
        MT4ORDERS::GetHistoryPositionData(Ticket);
      else
        MT4ORDERS::GetHistoryOrderData(-Ticket);
    }

    return(Res);
  }

  //           
  static bool SelectByPos( const int Index )
  {
    const int Total = ::PositionsTotal();
    const bool Flag = (Index < Total);

    const bool Res = (Flag) ? ::PositionSelectByTicket(::PositionGetTicket(Index)) : ::OrderSelect(::OrderGetTicket(Index - Total));

    if (Res)
    {
      if (Flag)
        MT4ORDERS::GetPositionData();
      else
        MT4ORDERS::GetOrderData();
    }

    return(Res);
  }

  static bool SelectByHistoryTicket( const int Ticket )
  {
    bool Res = ::HistoryDealSelect(Ticket) ? MT4HISTORY::IsMT4Deal(Ticket) : false;

    if (Res)
      MT4ORDERS::GetHistoryPositionData(Ticket);
    else
    {
      Res = ::HistoryOrderSelect(Ticket) ? MT4HISTORY::IsMT4Order(Ticket) : false;

      if (Res)
        MT4ORDERS::GetHistoryOrderData(Ticket);
    }

    return(Res);
  }

  static bool SelectByExistingTicket( const int Ticket )
  {
    bool Res = true;

    if (::PositionSelectByTicket(Ticket))
      MT4ORDERS::GetPositionData();
    else if (::OrderSelect(Ticket))
      MT4ORDERS::GetOrderData();
    else
      Res = false;

    return(Res);
  }

  //        :
  // MODE_TRADES:    >   >  >  
  // MODE_HISTORY:  >   >   >  
  static bool SelectByTicket( const int Ticket, const int Pool = MODE_TRADES )
  {
    return((Pool == MODE_TRADES) ?
           (MT4ORDERS::SelectByExistingTicket(Ticket) ? true : MT4ORDERS::SelectByHistoryTicket(Ticket)) :
           (MT4ORDERS::SelectByHistoryTicket(Ticket) ? true : MT4ORDERS::SelectByExistingTicket(Ticket)));
  }

public:
  static uint OrderSend_MaxPause; //      .

  static MqlTradeResult LastTradeResult;
  static MqlTradeRequest LastTradeRequest;
  static MqlTradeCheckResult LastTradeCheckResult;

  static bool MT4OrderSelect( const int Index, const int Select, const int Pool = MODE_TRADES )
  {
    return((Select == SELECT_BY_POS) ?
           ((Pool == MODE_TRADES) ? MT4ORDERS::SelectByPos(Index) : MT4ORDERS::SelectByPosHistory(Index)) :
           MT4ORDERS::SelectByTicket(Index, Pool));
  }

  //       MT5- OrderSelect
  static bool MT4OrderSelect( const ulong Ticket )
  {
    return(::OrderSelect(Ticket));
  }

  static int MT4OrdersTotal( void )
  {
    return(::OrdersTotal() + ::PositionsTotal());
  }

  //  ""     MT5- OrderSelect
  static int MT4OrdersTotal( const bool MT5 )
  {
    return(::OrdersTotal());
  }

  static int MT4OrdersHistoryTotal( void )
  {
    return(MT4ORDERS::History.GetAmount());
  }

  static int MT4OrderSend( const string Symb, const int Type, const double dVolume, const double Price, const int SlipPage, const double SL, const double TP,
                           const string comment = NULL, const long magic = 0, const datetime dExpiration = 0, color arrow_color = clrNONE )

  {
    ::ZeroMemory(MT4ORDERS::LastTradeRequest);

    MT4ORDERS::LastTradeRequest.action = (((Type == OP_BUY) || (Type == OP_SELL)) ? TRADE_ACTION_DEAL : TRADE_ACTION_PENDING);
    MT4ORDERS::LastTradeRequest.magic = magic;

    MT4ORDERS::LastTradeRequest.symbol = ((Symb == NULL) ? ::Symbol() : Symb);
    MT4ORDERS::LastTradeRequest.volume = dVolume;
    MT4ORDERS::LastTradeRequest.price = Price;

    MT4ORDERS::LastTradeRequest.tp = TP;
    MT4ORDERS::LastTradeRequest.sl = SL;
    MT4ORDERS::LastTradeRequest.deviation = SlipPage;
    MT4ORDERS::LastTradeRequest.type = (ENUM_ORDER_TYPE)Type;

    MT4ORDERS::LastTradeRequest.type_filling = MT4ORDERS::GetFilling(MT4ORDERS::LastTradeRequest.symbol, (uint)MT4ORDERS::LastTradeRequest.deviation);

    if (MT4ORDERS::LastTradeRequest.action == TRADE_ACTION_PENDING)
    {
      MT4ORDERS::LastTradeRequest.type_time = MT4ORDERS::GetExpirationType(MT4ORDERS::LastTradeRequest.symbol, (uint)dExpiration);

      if (dExpiration > ORDER_TIME_DAY)
        MT4ORDERS::LastTradeRequest.expiration = dExpiration;
    }

    MT4ORDERS::LastTradeRequest.comment = comment;

    return((arrow_color == INT_MAX) ? (MT4ORDERS::NewOrderCheck() ? 0 : -1) :
           ((((int)arrow_color != INT_MIN) || MT4ORDERS::NewOrderCheck()) &&
            MT4ORDERS::OrderSend(MT4ORDERS::LastTradeRequest, MT4ORDERS::LastTradeResult) ?
            (MT4ORDERS::IsHedging ? (int)MT4ORDERS::LastTradeResult.order : // PositionID == Result.order -  MT5-Hedge
             ((MT4ORDERS::LastTradeRequest.action == TRADE_ACTION_DEAL) ?
              (MT4ORDERS::IsTester ? (::PositionSelect(MT4ORDERS::LastTradeRequest.symbol) ? (int)PositionGetInteger(POSITION_TICKET) : 0) :
                                     (::HistoryDealSelect(MT4ORDERS::LastTradeResult.deal) ?
                                      (int)::HistoryDealGetInteger(MT4ORDERS::LastTradeResult.deal, DEAL_POSITION_ID) : 0)) :
              (int)MT4ORDERS::LastTradeResult.order)) : -1));
  }

  static bool MT4OrderModify( const int Ticket, const double Price, const double SL, const double TP, const datetime Expiration, const color Arrow_Color = clrNONE )
  {
    ::ZeroMemory(MT4ORDERS::LastTradeRequest);

               //  ,           
    bool Res = ((Ticket != MT4ORDERS::Order.Ticket) || (MT4ORDERS::Order.Ticket <= OP_SELL)) ?
               (MT4ORDERS::ModifyPosition(Ticket, MT4ORDERS::LastTradeRequest) ? true : MT4ORDERS::ModifyOrder(Ticket, Price, Expiration, MT4ORDERS::LastTradeRequest)) :
               (MT4ORDERS::ModifyOrder(Ticket, Price, Expiration, MT4ORDERS::LastTradeRequest) ? true : MT4ORDERS::ModifyPosition(Ticket, MT4ORDERS::LastTradeRequest));

//    if (Res) //   -  OrderCheck
    {
      MT4ORDERS::LastTradeRequest.tp = TP;
      MT4ORDERS::LastTradeRequest.sl = SL;

      Res = MT4ORDERS::NewOrderSend(Arrow_Color);
    }

    return(Res);
  }

  static bool MT4OrderClose( const int Ticket, const double dLots, const double Price, const int SlipPage, const color Arrow_Color = clrNONE )
  {
    //  MT4ORDERS::LastTradeRequest  MT4ORDERS::LastTradeResult,     ,    PositionGetString 
    ::PositionSelectByTicket(Ticket);

    ::ZeroMemory(MT4ORDERS::LastTradeRequest);

    MT4ORDERS::LastTradeRequest.action = TRADE_ACTION_DEAL;
    MT4ORDERS::LastTradeRequest.position = Ticket;

    MT4ORDERS::LastTradeRequest.symbol = ::PositionGetString(POSITION_SYMBOL);

    MT4ORDERS::LastTradeRequest.volume = dLots;
    MT4ORDERS::LastTradeRequest.price = Price;

    //    SL/TP-   .  -  
    MT4ORDERS::LastTradeRequest.tp = ::PositionGetDouble(POSITION_SL);
    MT4ORDERS::LastTradeRequest.sl = ::PositionGetDouble(POSITION_TP);

    MT4ORDERS::LastTradeRequest.deviation = SlipPage;

    MT4ORDERS::LastTradeRequest.type = (ENUM_ORDER_TYPE)(1 - ::PositionGetInteger(POSITION_TYPE));

    MT4ORDERS::LastTradeRequest.type_filling = MT4ORDERS::GetFilling(MT4ORDERS::LastTradeRequest.symbol, (uint)MT4ORDERS::LastTradeRequest.deviation);

    return(MT4ORDERS::NewOrderSend(Arrow_Color));
  }
  
  static bool MT4OrderCloseBy( const int Ticket, const int Opposite, const color Arrow_Color = clrNONE )
  {
    #ifdef _VERBOSE_
    Print(__FUNCSIG__, " ", Ticket, " ", Opposite);
    #endif
    if(!::PositionSelectByTicket(Opposite))
    {
      #ifdef _VERBOSE_
      Print("Can't select opposite ticket in OrderCloseBy:", Opposite);
      #endif
      return false;
    }
    
    int epyt = ::PositionGetInteger(POSITION_TYPE);
    
    //  MT4ORDERS::LastTradeRequest  MT4ORDERS::LastTradeResult,     ,    PositionGetString 
    if(!::PositionSelectByTicket(Ticket))
    {
      #ifdef _VERBOSE_
      Print("Can't select direct ticket in OrderCloseBy:", Ticket);
      #endif
      return false;
    }
    
    int type = ::PositionGetInteger(POSITION_TYPE);
    if(type == epyt)
    {
      #ifdef _VERBOSE_
      Print("Position types are the same in OrderCloseBy:", type);
      #endif
      return false;
    }

    ::ZeroMemory(MT4ORDERS::LastTradeRequest);

    MT4ORDERS::LastTradeRequest.action = TRADE_ACTION_CLOSE_BY;
    MT4ORDERS::LastTradeRequest.position = Ticket;
    MT4ORDERS::LastTradeRequest.position_by = Opposite;

    MT4ORDERS::LastTradeRequest.symbol = ::PositionGetString(POSITION_SYMBOL);

    MT4ORDERS::LastTradeRequest.type = ORDER_TYPE_CLOSE_BY;

    return(MT4ORDERS::NewOrderSend(Arrow_Color));
  }
  

  static bool MT4OrderDelete( const int Ticket, const color Arrow_Color = clrNONE )
  {
//    bool Res = ::OrderSelect(Ticket); //   ,   MT4ORDERS::LastTradeRequest  MT4ORDERS::LastTradeResult ?

    ::ZeroMemory(MT4ORDERS::LastTradeRequest);

    MT4ORDERS::LastTradeRequest.action = TRADE_ACTION_REMOVE;
    MT4ORDERS::LastTradeRequest.order = Ticket;

    if ((!MT4ORDERS::IsTester) && ::OrderSelect(Ticket)) //   MT4ORDERS::SymbolTrade()
      MT4ORDERS::LastTradeRequest.symbol = ::OrderGetString(ORDER_SYMBOL);

    return(MT4ORDERS::NewOrderSend(Arrow_Color));
  }

#define MT4_ORDERFUNCTION(NAME,T)  \
  static T MT4Order##NAME( void )  \
  {                                \
    return(MT4ORDERS::Order.NAME); \
  }

  MT4_ORDERFUNCTION(Ticket, int)
  MT4_ORDERFUNCTION(Type, int)
  MT4_ORDERFUNCTION(Lots, double)
  MT4_ORDERFUNCTION(OpenPrice, double)
  MT4_ORDERFUNCTION(OpenTime, datetime)
  MT4_ORDERFUNCTION(StopLoss, double)
  MT4_ORDERFUNCTION(TakeProfit, double)
  MT4_ORDERFUNCTION(ClosePrice, double)
  MT4_ORDERFUNCTION(CloseTime, datetime)
  MT4_ORDERFUNCTION(Expiration, datetime)
  MT4_ORDERFUNCTION(MagicNumber, long)
  MT4_ORDERFUNCTION(Profit, double)
//  MT4_ORDERFUNCTION(Commission, double)
  MT4_ORDERFUNCTION(Swap, double)
  MT4_ORDERFUNCTION(Symbol, string)
//  MT4_ORDERFUNCTION(Comment, string)

#undef MT4_ORDERFUNCTION

  static double MT4OrderCommission( void )
  {
    MT4ORDERS::CheckPositionCommissionComment();

    return(MT4ORDERS::Order.Commission);
  }

  static string MT4OrderComment( void )
  {
    MT4ORDERS::CheckPositionCommissionComment();

    return(MT4ORDERS::Order.Comment);

  }

  static void MT4OrderPrint( void )
  {
    MT4ORDERS::CheckPositionCommissionComment();

    ::Print(MT4ORDERS::Order.ToString());

    return;
  }
/*
  static string MT4OrderToString( void )
  {
    return(MT4ORDERS::Order.ToString());
  } */
};

// #define OrderToString MT4ORDERS::MT4OrderToString

static MT4_ORDER MT4ORDERS::Order = {0};

static MT4HISTORY MT4ORDERS::History;

static const bool MT4ORDERS::IsTester = (::MQLInfoInteger(MQL_TESTER) || ::MQLInfoInteger(MQL_OPTIMIZATION) ||
                                         ::MQLInfoInteger(MQL_VISUAL_MODE) || ::MQLInfoInteger(MQL_FRAME_MODE));

static const bool MT4ORDERS::IsHedging = ((ENUM_ACCOUNT_MARGIN_MODE)::AccountInfoInteger(ACCOUNT_MARGIN_MODE) ==
                                          ACCOUNT_MARGIN_MODE_RETAIL_HEDGING) && MT4ORDERS::IsTester;

static bool MT4ORDERS::OrderSendBug = false;

static uint MT4ORDERS::OrderSend_MaxPause = 1000000; //      .

static MqlTradeResult MT4ORDERS::LastTradeResult = {0};
static MqlTradeRequest MT4ORDERS::LastTradeRequest = {0};
static MqlTradeCheckResult MT4ORDERS::LastTradeCheckResult = {0};

bool OrderClose( const int Ticket, const double dLots, const double Price, const int SlipPage, const color Arrow_Color = clrNONE )
{
  return(MT4ORDERS::MT4OrderClose(Ticket, dLots, Price, SlipPage, Arrow_Color));
}

bool OrderCloseBy( const int Ticket, const int Opposite, const color Arrow_Color = clrNONE )
{
  return(MT4ORDERS::MT4OrderCloseBy(Ticket, Opposite, Arrow_Color));
}

bool OrderModify( const int Ticket, const double Price, const double SL, const double TP, const datetime Expiration, const color Arrow_Color = clrNONE )
{
  return(MT4ORDERS::MT4OrderModify(Ticket, Price, SL, TP, Expiration, Arrow_Color));
}

bool OrderDelete( const int Ticket, const color Arrow_Color = clrNONE )
{
  return(MT4ORDERS::MT4OrderDelete(Ticket, Arrow_Color));
}

void OrderPrint( void )
{
  MT4ORDERS::MT4OrderPrint();

  return;
}

#define MT4_ORDERGLOBALFUNCTION(NAME,T)  \
  T Order##NAME( void )                  \
  {                                      \
    return(MT4ORDERS::MT4Order##NAME()); \
  }

MT4_ORDERGLOBALFUNCTION(sHistoryTotal, int)
MT4_ORDERGLOBALFUNCTION(Ticket, int)
MT4_ORDERGLOBALFUNCTION(Type, int)
MT4_ORDERGLOBALFUNCTION(Lots, double)
MT4_ORDERGLOBALFUNCTION(OpenPrice, double)
MT4_ORDERGLOBALFUNCTION(OpenTime, datetime)
MT4_ORDERGLOBALFUNCTION(StopLoss, double)
MT4_ORDERGLOBALFUNCTION(TakeProfit, double)
MT4_ORDERGLOBALFUNCTION(ClosePrice, double)
MT4_ORDERGLOBALFUNCTION(CloseTime, datetime)
MT4_ORDERGLOBALFUNCTION(Expiration, datetime)
MT4_ORDERGLOBALFUNCTION(MagicNumber, long)
MT4_ORDERGLOBALFUNCTION(Profit, double)
MT4_ORDERGLOBALFUNCTION(Commission, double)
MT4_ORDERGLOBALFUNCTION(Swap, double)
MT4_ORDERGLOBALFUNCTION(Symbol, string)
MT4_ORDERGLOBALFUNCTION(Comment, string)

#undef MT4_ORDERGLOBALFUNCTION

//   
#define OrdersTotal MT4ORDERS::MT4OrdersTotal //  Expert/Expert.mqh -   MT5-OrdersTotal()

bool OrderSelect( const int Index, const int Select, const int Pool = MODE_TRADES )
{
  return(MT4ORDERS::MT4OrderSelect(Index, Select, Pool));
}

int OrderSend( const string Symb, const int Type, const double dVolume, const double Price, const int SlipPage, const double SL, const double TP,
               const string comment = NULL, const long magic = 0, const datetime dExpiration = 0, color arrow_color = clrNONE )
{
  return(MT4ORDERS::MT4OrderSend(Symb, Type, dVolume, Price, SlipPage, SL, TP, comment, magic, dExpiration, arrow_color));
}

#define RETURN_ASYNC(A) return((A) && ::OrderSendAsync(MT4ORDERS::LastTradeRequest, MT4ORDERS::LastTradeResult) &&                        \
                               (MT4ORDERS::LastTradeResult.retcode == TRADE_RETCODE_PLACED) ? MT4ORDERS::LastTradeResult.request_id : 0);

uint OrderCloseAsync( const int Ticket, const double dLots, const double Price, const int SlipPage, const color Arrow_Color = clrNONE )
{
  RETURN_ASYNC(OrderClose(Ticket, dLots, Price, SlipPage, INT_MAX))
}

uint OrderModifyAsync( const int Ticket, const double Price, const double SL, const double TP, const datetime Expiration, const color Arrow_Color = clrNONE )
{
  RETURN_ASYNC(OrderModify(Ticket, Price, SL, TP, Expiration, INT_MAX));
}

uint OrderDeleteAsync( const int Ticket, const color Arrow_Color = clrNONE )
{
  RETURN_ASYNC(OrderDelete(Ticket, INT_MAX));
}

uint OrderSendAsync( const string Symb, const int Type, const double dVolume, const double Price, const int SlipPage, const double SL, const double TP,
                    const string comment = NULL, const long magic = 0, const datetime dExpiration = 0, color arrow_color = clrNONE )
{
  RETURN_ASYNC(!OrderSend(Symb, Type, dVolume, Price, SlipPage, SL, TP, comment, magic, dExpiration, INT_MAX));
}

#undef RETURN_ASYNC

#endif // __MT4ORDERS__
#endif // __MQL5__